# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from logging import getLogger
import re

logger = getLogger(__name__)

ILLEGAL_XML_RE = re.compile(
    '[\x00-\x08\x0b-\x1f\x7f-\x84\x86-\x9f\ud800-\udfff\ufdd0-\ufddf'
    '\ufffe-\uffff]')


def safe_utf8(text):
    return ILLEGAL_XML_RE.sub(' ', text)



# -*- coding: utf-8 -*-
"""
Batik-based rasterizer module.

Download the latest batik rasterizer to use the module. Note Ubuntu 16.04LTS
package is broken and does not work.

Prerequisite:

    wget http://www.apache.org/dyn/mirrors/mirrors.cgi?action=download&\
    filename=xmlgraphics/batik/binaries/batik-bin-1.9.tar.gz
    tar xzf batik-bin-1.9.tar.gz
    export BATIK_PATH=./batik-bin-1.9

Deb package:

    sudo apt-get install -y libbatik-java


"""
from __future__ import absolute_import, unicode_literals

from PIL import Image
import logging
import os
import subprocess
from psd2svg.utils import temporary_directory
from psd2svg.rasterizer.base_rasterizer import BaseRasterizer

logger = logging.getLogger(__name__)

BATIK_PATH = os.environ.get(
    'BATIK_PATH', "/usr/share/java/batik-rasterizer.jar"
)

class BatikRasterizer(BaseRasterizer):
    """Batik rasterizer."""

    def __init__(self, jar_path=None, **kwargs):
        self.jar_path = jar_path if jar_path else BATIK_PATH
        assert os.path.exists(self.jar_path)

    def rasterize(self, url, size=None, format="png"):
        with temporary_directory() as d:
            basename, ext = os.path.splitext(os.path.basename(url))
            output_file = os.path.join(d, "{}.{}".format(basename, format))
            cmd = [
                "java", "-Djava.awt.headless=true",
                "-jar", self.jar_path,
                "-bg", "0.255.255.255",
                "-m", "image/{}".format(format),
                "-d", d,
                "{}".format(url),
            ]
            if size:
                cmd += ["-w", size[0], "-h", size[1]]
            proc = subprocess.Popen(
                cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE
            )
            stdout, stderr = proc.communicate()
            try:
                assert os.path.exists(output_file)
                rasterized = Image.open(output_file)
            except:
                logger.error("{}\n{}{}".format(" ".join(cmd), stdout, stderr))
                raise
            return self.composite_background(rasterized)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import logging
import tempfile
import contextlib
import shutil

logger = logging.getLogger(__file__)


@contextlib.contextmanager
def temporary_directory(*args, **kwargs):
    """For Python 2.x compatibility."""
    d = tempfile.mkdtemp(*args, **kwargs)
    try:
        yield d
    finally:
        shutil.rmtree(d)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
from logging import getLogger
import svgwrite
from psd_tools.user_api import BBox
from psd_tools.constants import TaggedBlock
from psd2svg.converter.constants import BLEND_MODE
from psd2svg.utils.xml import safe_utf8
from psd2svg.utils.color import cmyk2rgb
import numpy as np

logger = getLogger(__name__)


class LayerConverter(object):

    def convert_layer(self, layer):
        """
        Convert the given layer.

        The current implementation always converts a PSD layer to a single
        SVG element.

        :return: SVG element.
        """
        if layer.is_group():
            element = self.create_group(layer)

        elif layer.kind == 'shape':
            if layer.has_path():
                element = self.create_path(layer)
            else:
                element = self.create_rect(layer)
            # TODO: Deal with coflict in add_attributes() later.
            # Blending, mask, and class names conflict.
            element = self.add_stroke_style(layer, element)
            element = self.add_stroke_content_style(layer, element)

        elif layer.has_pixels():
            element = self.create_image(layer)
            if layer.kind == 'type':
                # TODO: Embed text metadata.
                # text = self._get_text(layer)
                pass

        elif layer.kind == 'adjustment':
            element = self.create_adjustment(layer)
            # TODO: Wrap previous elements with the adjustment.

        else:
            # Boxless shape fill.
            element = self.create_rect(layer)

        element = self.add_fill(layer, element)
        element = self.add_attributes(layer, element)
        if layer.has_mask():
            mask = layer.mask
            if mask.has_box() and not mask.disabled and (
                not mask.user_mask_from_render or layer.has_vector_mask()):
                mask_element = self.create_mask(layer)
                element['mask'] = mask_element.get_funciri()

        element = self.add_effects(layer, element)
        return element

    def create_group(self, group, element=None):
        """Create and fill in a new group element."""
        if not element:
            element = self._dwg.g()
        for child_layer in reversed(group.layers):
            child_element = self.convert_layer(child_layer)
            element.add(child_element)

            # Clipping.
            if len(child_layer.clip_layers) > 0:
                clip_group = self.create_clipping(child_layer, child_element)
                element.add(clip_group)

                # TODO: Blending option.
                # layer.get_tag(TaggedBlock.BLEND_CLIPPING_ELEMENTS)
        return element

    def create_image(self, layer):
        """Create an image element."""
        element = self._dwg.image(
            self._get_image_href(layer.as_PIL()),
            insert=(layer.left, layer.top),
            size=(layer.width, layer.height),
            debug=False)  # To disable attribute validation.
        return element

    def create_path(self, layer):
        """Create a path element."""
        path = self._dwg.path(d=self.generate_path(layer.vector_mask))
        if layer.vector_mask.initial_fill_rule:
            element = self._dwg.defs.add(self._dwg.rect(
                insert=(self._psd.bbox.x1, self._psd.bbox.y1),
                size=(self._psd.bbox.width, self._psd.bbox.height)))
            mask = self._dwg.defs.add(self._dwg.mask())
            mask.add(self._dwg.rect(
                insert=(self._psd.bbox.x1, self._psd.bbox.y1),
                size=(self._psd.bbox.width, self._psd.bbox.height),
                fill='white'))
            path['fill'] = 'black'
            path['clip-rule'] = 'evenodd'
            mask.add(path)
            element['mask'] = mask.get_funciri()
            use = self._dwg.use(element.get_iri())
            return use
        else:
            path['fill-rule'] = 'evenodd'
            return path

    def create_rect(self, layer):
        """Create a shape or adjustment element."""
        if layer.has_box():
            element = self._dwg.rect(
                insert=(layer.left, layer.top),
                size=(layer.width, layer.height))
        else:
            element = self._dwg.rect(
                insert=(self._psd.bbox.x1, self._psd.bbox.y1),
                size=(self._psd.bbox.width, self._psd.bbox.height))
        element['fill'] = 'none'
        return element

    def create_mask(self, layer):
        """Create a mask."""
        if not layer.has_mask() or not layer.mask.has_box():
            return None

        # In SVG, mask needs a default rect.
        viewbox = layer.bbox
        if viewbox.is_empty():
            viewbox = BBox(0, 0, self.width, self.height)
        mask_element = self._dwg.defs.add(self._dwg.mask(
            size=(viewbox.width, viewbox.height)))
        mask_element.add(self._dwg.rect(
            insert=(viewbox.x1, viewbox.y1),
            size=(viewbox.width, viewbox.height),
            fill='rgb({0},{0},{0})'.format(layer.mask.background_color)))
        mask_element.add(self._dwg.image(
            self._get_image_href(layer.mask.as_PIL()),
            size=(layer.mask.width, layer.mask.height),
            insert=(layer.mask.left, layer.mask.top)))
        mask_element['color-interpolation'] = 'sRGB'
        return mask_element

    def create_clipping(self, layer, clip_element):
        """Create clipped elements."""
        # Create a mask for this clip element.
        if isinstance(clip_element, svgwrite.path.Path):
            clippath = self._dwg.defs.add(self._dwg.clipPath())
            use = self._dwg.use(clip_element.get_iri())
            clippath.add(use)
            element = self._dwg.g()
            element['clip-path'] = clippath.get_funciri()
        else:
            mask = self._dwg.defs.add(self._dwg.mask())
            use = self._dwg.use(clip_element.get_iri())
            use['filter'] = self._get_white_filter().get_funciri()
            mask.add(use)
            mask['color-interpolation'] = 'sRGB'
            element = self._dwg.g(mask=mask.get_funciri())

        element['class'] = 'psd-clipping'
        for child_layer in reversed(layer.clip_layers):
            clipped_element = self.convert_layer(child_layer)
            if clipped_element:
                element.add(clipped_element)
        return element

    def create_preview(self, hidden=True):
        """Create a preview image of the photoshop."""
        element = self._dwg.image(
            self._get_image_href(self._psd.as_PIL()),
            insert=(0, 0),
            size=(self.width, self.height))
        element['class'] = 'photoshop-image'
        if hidden:
            element['visibility'] = 'hidden'
        return element

    def add_fill(self, layer, element):
        """Add fill attribute to the given element."""
        if layer.has_tag(TaggedBlock.SOLID_COLOR_SHEET_SETTING):
            effect = layer.get_tag(TaggedBlock.SOLID_COLOR_SHEET_SETTING)
            element['fill'] = self.create_solid_color(effect)
        elif layer.has_tag(TaggedBlock.PATTERN_FILL_SETTING):
            effect = layer.get_tag(TaggedBlock.PATTERN_FILL_SETTING)
            pattern_element = self.create_pattern(effect)
            element['fill'] = pattern_element.get_funciri()
        elif layer.has_tag(TaggedBlock.GRADIENT_FILL_SETTING):
            effect = layer.get_tag(TaggedBlock.GRADIENT_FILL_SETTING)
            gradient = self.create_gradient(
                effect, (layer.width, layer.height))
            element['fill'] = gradient.get_funciri()
        return element

    def add_attributes(self, layer, element):
        """Add layer attributes such as blending or visibility options."""
        element.set_desc(title=safe_utf8(layer.name))
        element['class'] = 'psd-layer psd-{}'.format(layer.kind)
        if not layer.visible:
            element['visibility'] = 'hidden'
        if layer.opacity < 255.0:
            element['opacity'] = layer.opacity / 255.0
        self.add_blend_mode(element, layer.blend_mode)
        return element

    def add_blend_mode(self, element, blend_mode):
        """Set blending option to the element."""
        blend_mode = BLEND_MODE.get(blend_mode, 'normal')
        if blend_mode != 'normal':
            if 'style' in element.attribs:
                element['style'] += 'mix-blend-mode: {};'.format(blend_mode)
            else:
                element['style'] = 'mix-blend-mode: {};'.format(blend_mode)

    def create_solid_color(self, effect):
        """
        Create a fill attribute.

        This is supposed to be solidColor of SVG 1.2 Tiny spec, but for now,
        implement as fill attribute.

        :rtype: str
        """
        if hasattr(effect, 'color'):
            color = effect.color
        else:
            color = effect
        if color.name == 'rgb':
            return 'rgb({},{},{})'.format(*map(int, color.value))
        elif color.name == 'gray':
            return 'rgb({0},{0},{0})'.format(int(255 * color.value[0]))
        elif color.name == 'cmyk':
            rgb = cmyk2rgb(color.value)
            return 'rgb({},{},{})'.format(*map(int, rgb))
        else:
            logger.warning('Unsupported color: {}'.format(color))
            return 'rgba(0,0,0,0)'

    def create_pattern(self, effect, insert=(0, 0)):
        """Create a pattern element."""
        pattern = self._psd.patterns.get(effect.pattern.id)
        phase = effect.phase
        scale = effect.scale.value  # TODO: Check unit
        if not pattern:
            logger.error('Pattern data not found')
            return self._dwg.defs.add(svgwrite.pattern.Pattern())

        element = self._dwg.defs.add(svgwrite.pattern.Pattern(
            width=pattern.width,
            height=pattern.height,
            patternUnits='userSpaceOnUse',
            patternContentUnits='userSpaceOnUse',
            patternTransform='translate({},{}) scale({})'.format(
                insert[0] + phase[0], insert[1] + phase[1], scale / 100.0),
        ))
        element.add(self._dwg.image(
            self._get_image_href(pattern.as_PIL()),
            insert=(0, 0),
            size=(pattern.width, pattern.height),
        ))
        return element

    def create_gradient(self, effect, size):
        if effect.type == 'radial':
            element = self._dwg.defs.add(svgwrite.gradients.RadialGradient(
                center=None, r=.5))
        else:
            theta = np.radians(-effect.angle.value)
            start = np.array([size[0] * np.cos(theta - np.pi),
                              size[1] * np.sin(theta - np.pi)])
            end = np.array([size[0] * np.cos(theta), size[1] * np.sin(theta)])
            r = 1.0 * np.max(np.abs(start))

            start = start / (2.0 * r) + 0.5
            end = end / (2.0 * r) + 0.5

            start = [np.around(x, decimals=6) for x in start]
            end = [np.around(x, decimals=6) for x in end]

            element = self._dwg.defs.add(svgwrite.gradients.LinearGradient(
                start=start, end=end))

        gradient = effect.gradient
        if not gradient.colors:
            logger.warning("Unsupported gradient type: {}".format(gradient))
            return element

        # Interpolate color and opacity for both points.
        cp = np.array([x.location / 4096.0 for x in gradient.colors])
        op = np.array([x.location / 4096.0 for x in gradient.transform])
        c_items = np.array([
            x.color.value for x in gradient.colors]).transpose()
        o_items = np.array([x.opacity.value / 100.0
                            for x in gradient.transform])  # TODO: Check unit.

        # Remove duplicate points.
        index = np.concatenate((np.diff(cp) > 0, [True]))
        if np.any(np.logical_not(index)):
            logger.warning('Duplicate gradient color stop: {}'.format(cp))
        cp = cp[index]
        c_items = c_items[:, index]
        index = np.concatenate((np.diff(op) > 0, [True]))
        if np.any(np.logical_not(index)):
            logger.warning('Duplicate gradient opacity stop: {}'.format(op))
        op = op[index]
        o_items = o_items[index]

        # Single point handling.
        if len(cp) < 2:
            cp = np.array([0.0, 1.0])
            c_items = np.concatenate((c_items, c_items), axis=1)
        if len(op) < 2:
            op = np.array([0.0, 1.0])
            o_items = np.array(list(o_items) + list(o_items))

        # Reverse if specified.
        if effect.reversed:
            cp = 1.0 - cp[::-1]
            op = 1.0 - op[::-1]
            c_items = c_items[:, ::-1]
            o_items = o_items[::-1]

        mp = np.unique(np.concatenate((cp, op)))
        fc = np.stack(
            [np.interp(mp, cp, c_items[index, :]) for index in range(3)])
        fo = np.interp(mp, op, o_items)

        for index in range(len(mp)):
            color = tuple(fc[:, index].astype(np.uint8).tolist())
            element.add_stop_color(offset=mp[index], opacity=fo[index],
                                   color='rgb{}'.format(color))
        return element



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import logging
import tempfile
from PIL import Image

logger = logging.getLogger(__name__)


class BaseRasterizer(object):
    """Base class for rasterizer implementation."""

    def rasterize_from_string(self, input, **kwargs):
        with tempfile.NamedTemporaryFile(suffix='.svg') as f:
            f.write(input if isinstance(input, bytes)
                    else input.encode('utf-8'))
            f.flush()
            return self.rasterize(f.name, **kwargs)

    def rasterize(self, url, size=None, **kwargs):
        raise NotImplementedError

    def composite_background(self, image):
        background = Image.new(
            "RGBA", size=image.size, color=(255, 255, 255, 0))
        background.alpha_composite(image)
        return background



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
from logging import getLogger

logger = getLogger(__name__)


def cmyk2rgb(cmyk):
    return (2.55 * (1.0 - cmyk[0]) * (1.0 - cmyk[3]),
            2.55 * (1.0 - cmyk[1]) * (1.0 - cmyk[3]),
            2.55 * (1.0 - cmyk[2]) * (1.0 - cmyk[3]))



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from psd_tools.constants import BlendMode, BlendMode2

# https://helpx.adobe.com/photoshop/using/blending-modes.html
BLEND_MODE = {
    # Layer modes.
    BlendMode.PASS_THROUGH: 'normal /*pass*/',
    BlendMode.NORMAL: 'normal',
    BlendMode.DISSOLVE: 'normal /*dissolve*/',
    BlendMode.DARKEN: 'darken',
    BlendMode.MULTIPLY: 'multiply',
    BlendMode.COLOR_BURN: 'color-burn',
    BlendMode.LINEAR_BURN: 'plus-darker /*linear-burn*/',  # Only webkit
    BlendMode.DARKER_COLOR: 'darken /*darker-color*/',
    BlendMode.LIGHTEN: 'lighten',
    BlendMode.SCREEN: 'screen',
    BlendMode.COLOR_DODGE: 'color-dodge',
    BlendMode.LINEAR_DODGE: 'plus-lighter /*linear-dodge*/',  # Only webkit
    BlendMode.LIGHTER_COLOR: 'lighten /*lighter-color*/',
    BlendMode.OVERLAY: 'overlay',
    BlendMode.SOFT_LIGHT: 'soft-light',
    BlendMode.HARD_LIGHT: 'hard-light',
    BlendMode.VIVID_LIGHT: 'lighten /*vivid-light*/',
    BlendMode.LINEAR_LIGHT: 'darken /*linear-light*/',
    BlendMode.PIN_LIGHT: 'normal /*pin-light*/',
    BlendMode.HARD_MIX: 'normal /*hard-mix*/',
    BlendMode.DIFFERENCE: 'difference',
    BlendMode.EXCLUSION: 'exclusion',
    BlendMode.SUBTRACT: 'difference /*subtract*/',
    BlendMode.DIVIDE: 'difference /*divide*/',
    BlendMode.HUE: 'hue',
    BlendMode.SATURATION: 'saturation',
    BlendMode.COLOR: 'color',
    BlendMode.LUMINOSITY: 'luminosity',
    # Descriptor-based modes.
    BlendMode2.NORMAL: 'normal',
    BlendMode2.DISSOLVE: 'normal /*dissolve*/',
    BlendMode2.DARKEN: 'darken',
    BlendMode2.MULTIPLY: 'multiply',
    BlendMode2.COLOR_BURN: 'color-burn',
    BlendMode2.LINEAR_BURN: 'plus-darker /*linear-burn*/',  # Only webkit
    BlendMode2.DARKER_COLOR: 'darken /*darker-color*/',
    BlendMode2.LIGHTEN: 'lighten',
    BlendMode2.SCREEN: 'screen',
    BlendMode2.COLOR_DODGE: 'color-dodge',
    BlendMode2.LINEAR_DODGE: 'plus-lighter /*linear-dodge*/',  # Only webkit
    BlendMode2.LIGHTER_COLOR: 'lighten /*lighter-color*/',
    BlendMode2.OVERLAY: 'overlay',
    BlendMode2.SOFT_LIGHT: 'soft-light',
    BlendMode2.HARD_LIGHT: 'hard-light',
    BlendMode2.VIVID_LIGHT: 'lighten /*vivid-light*/',
    BlendMode2.LINEAR_LIGHT: 'darken /*linear-light*/',
    BlendMode2.PIN_LIGHT: 'normal /*pin-light*/',
    BlendMode2.HARD_MIX: 'normal /*hard-mix*/',
    BlendMode2.DIFFERENCE: 'difference',
    BlendMode2.EXCLUSION: 'exclusion',
    BlendMode2.SUBTRACT: 'difference /*subtract*/',
    BlendMode2.DIVIDE: 'difference /*divide*/',
    BlendMode2.HUE: 'hue',
    BlendMode2.SATURATION: 'saturation',
    BlendMode2.COLOR: 'color',
    BlendMode2.LUMINOSITY: 'luminosity',
    # psd-tools API modes.
    'pass through': 'normal /*pass through*/',
    'dissolve': 'normal /*dissolve*/',
    'darken': 'darken',
    'multiply': 'multiply',
    'color burn': 'color-burn',
    'linear burn': 'plus-darker /*linear-burn*/',
    'darker color': 'darken /*darker-color*/',
    'lighten': 'lighten',
    'screen': 'screen',
    'color dodge': 'color-dodge',
    'linear dodge': 'plus-lighter /*linear-dodge*/',
    'lighter color': 'lighten /*lighter-color*/',
    'overlay': 'overlay',
    'soft light': 'soft-light',
    'hard light': 'hard-light',
    'vivid light': 'lighten /*vivid-light*/',
    'linear light': 'darken /*linear-light*/',
    'pin light': 'normal /*pin-light*/',
    'hard mix': 'normal /*hard-mix*/',
    'difference': 'difference',
    'exclusion': 'exclusion',
    'subtract': 'difference /*subtract*/',
    'divide': 'difference /*divide*/',
    'hue': 'hue',
    'saturation': 'saturation',
    'color': 'color',
    'luminosity': 'luminosity',
    # Descriptor values.
    'linearnormaldodge': 'plus-lighter /*linear-dodge*/',
    'linearnormalburn': 'plus-darker /*linear-burn*/',
}

JUSTIFICATIONS = {
    0: 'start',
    1: 'end',
    2: 'middle',
}



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
from logging import getLogger
from psd2svg.converter.constants import JUSTIFICATIONS
from psd2svg.utils.color import cmyk2rgb
from psd2svg.utils.xml import safe_utf8


logger = getLogger(__name__)


class TextConverter(object):

    def _get_text(self, layer):
        text_info = _get_text_info(layer)
        if not text_info:
            return None

        text = self._dwg.text('', insert=(0, 0))
        text['font-size'] = 0  # To discard whitespace between spans.
        text['text-anchor'] = 'middle'

        transform = text_info['matrix']
        if transform[1] != 0.0 and transform[2] != 0.0:
            text['transform'] = 'matrix{}'.format(transform)
        elif transform[0] != 1.0 and transform[3] != 1.0:
            text['transform'] = 'translate({},{}) scale({},{})'.format(
                transform[4], transform[5], transform[0], transform[3])
        else:
            text['transform'] = 'translate({},{})'.format(
                transform[4], transform[5])

        text['text-anchor'] = JUSTIFICATIONS.get(
            text_info['justification'], 'start')

        if text_info['direction']:
            text['writing-mode'] = 'tb'
            text['glyph-orientation-vertical'] = 0

        newline = False
        for span in text_info['spans']:
            rspans = span[b'Text'].split('\r')
            for index in range(len(rspans)):
                if index > 0:
                    newline = True
                if len(rspans[index]) == 0:
                    continue

                value = safe_utf8(rspans[index]).replace(u' ', '\u00a0')
                # Whitespace workaround, because newline is ignored.
                tspan = self._dwg.tspan(value)
                if newline:
                    if text_info['direction']:
                        tspan['y'] = 0
                        tspan['dx'] = '-1em'
                    else:
                        tspan['x'] = 0
                        tspan['dy'] = '1em'
                    newline = False

                tspan['font-family'] = span[b'Font'][b'Name']

                if int(span[b'Font'][b'Synthetic']) & 1:
                    tspan['font-style'] = 'italic'
                if int(span[b'Font'][b'Synthetic']) & 2:
                    tspan['font-weight'] = 'bold'

                fontsize = span.get(b'FontSize', 12)  # Not sure default 12...
                # SVG cannot apply per-letter scaling...
                if span.get(b'HorizontalScale', None) is not None:
                    fontsize *= span[b'HorizontalScale']
                tspan['font-size'] = fontsize
                if span.get(b'Tracking', None):
                    tspan['letter-spacing'] = span[b'Tracking'] / 100

                decoration = []
                if span.get(b'Underline', False):
                    decoration.append('underline')
                if span.get(b'Strikethrough', False):
                    decoration.append('line-through')
                if len(decoration) > 0:
                    tspan['text-decoration'] = " ".join(decoration)

                if b'FillColor' in span:
                    if span[b'FillColor'][b'Type'] == 0:
                        gray = span[b'FillColor'][b'Values'][1]
                        rgb = (gray, gray, gray)
                    elif span[b'FillColor'][b'Type'] == 1:
                        rgb = [int(c*255) for c in
                               span[b'FillColor'][b'Values'][1:]]
                        if len(rgb) != 3:
                            raise ValueError('Unsupported FillColor')
                    elif span[b'FillColor'][b'Type'] == 2:
                        cmyk = span[b'FillColor'][b'Values'][1:]
                        if len(cmyk) != 4:
                            raise ValueError('Unsupported FillColor')
                        rgb = [int(round(c * 100)) for c in cmyk2rgb(cmyk)]
                    else:
                        raise ValueError('Unsupported FillColor')
                    opacity = span[b'FillColor'][b'Values'][0]
                else:
                    rgb = (0, 0, 0)
                    opacity = 1.0
                tspan['fill'] = 'rgb({},{},{})'.format(*rgb)
                tspan['fill-opacity'] = opacity

                text.add(tspan)

        return text


def _get_text_info(layer):
    type_info = dict(layer._record.tagged_blocks).get(b'TySh', None)
    if type_info is None:
        return None
    engine_data = dict(type_info.text_data.items)[b'EngineData']
    fontset = engine_data[b'DocumentResources'][b'FontSet']
    direction = engine_data[b'EngineDict'][
        b'Rendered'][b'Shapes'][b'WritingDirection']
    # Matrix [xx xy yx yy tx ty] applies affine transformation.
    matrix = (type_info.xx, type_info.xy, type_info.yx, type_info.yy,
              type_info.tx, type_info.ty)

    paragraphs = engine_data[b'EngineDict'][b'ParagraphRun'][b'RunArray']
    justification = paragraphs[0][
        b'ParagraphSheet'][b'Properties'].get(b'Justification', 0)

    runlength = engine_data[b'EngineDict'][b'StyleRun'][b'RunLengthArray']
    runarray = engine_data[b'EngineDict'][b'StyleRun'][b'RunArray']
    text = engine_data[b'EngineDict'][b'Editor'][b'Text']

    start = 0
    spans = []
    for run, size in zip(runarray, runlength):
        runtext = text[start:start+size]
        stylesheet = run[b'StyleSheet'][b'StyleSheetData'].copy()
        stylesheet[b'Text'] = runtext
        stylesheet[b'Font'] = fontset[stylesheet.get(b'Font', 0)]
        spans.append(stylesheet)
        start += size
    return {'direction': direction,
            'spans': spans,
            'justification': justification,
            'matrix': matrix}



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
import argparse
import logging
import os
from psd2svg import psd2svg

def main():
    parser = argparse.ArgumentParser(description='Convert PSD file to SVG')
    parser.add_argument(
        'input', metavar='INPUT', type=str, help='Input PSD file path or URL')
    parser.add_argument(
        'output', metavar='PATH', type=str, nargs='?', default='.',
        help='Output file or directory. When directory is specified, filename'
             ' is automatically inferred from input')
    parser.add_argument(
        '--text-mode', metavar='MODE', default='image',
        help='Render texts in bitmap or vector. One of [image, image-only, '
             'text, text-only]. Default image')
    parser.add_argument(
        '--export-resource', action='store_true',
        help='Export image resources, default False')
    parser.add_argument(
        '--resource-prefix', metavar='PATH', type=str, default='.',
        help='Resource path prefix relative to output.')
    parser.add_argument(
        '--overwrite', action='store_true', help='Overwrite output')
    parser.add_argument(
        '--embed-preview', action='store_true', help='Embed a preview')
    parser.add_argument(
        '--rasterizer', metavar='METHOD', default='chromium', type=str,
        help='Specify which rasterizer to use. default chromium.')
    parser.add_argument(
        '--loglevel', metavar='LEVEL', default='WARNING',
        help='Logging level, default WARNING')
    args = parser.parse_args()

    logging.basicConfig(level=getattr(logging, args.loglevel.upper(),
                                      'WARNING'))

    prefix, ext = os.path.splitext(args.output)
    if ext.lower() in (".png", ".jpg", ".jpeg", ".gif" ".tiff"):
        from psd2svg.rasterizer import create_rasterizer
        rasterizer = create_rasterizer(args.rasterizer)
        svg_file = prefix + ".svg"
        psd2svg(args.input, svg_file,
                resource_prefix=args.resource_prefix,
                text_mode=args.text_mode,
                export_resource=args.export_resource,
                overwrite=args.overwrite,
                embed_preview=args.embed_preview)
        image = rasterizer.rasterize(svg_file)
        image.save(args.output)
    else:
        psd2svg(args.input, args.output,
                resource_prefix=args.resource_prefix,
                text_mode=args.text_mode,
                export_resource=args.export_resource,
                overwrite=args.overwrite,
                embed_preview=args.embed_preview)


if __name__ == '__main__':
    main()



from psd2svg.rasterizer import create_rasterizer
import logging

logger = logging.getLogger(__name__)


def main():
    import argparse

    parser = argparse.ArgumentParser(description="Rasterize SVG file.")
    parser.add_argument("input", metavar="URL", type=str, help="Input URL.")
    parser.add_argument("--output", metavar="PATH", default="output.png",
                        help="Output file. default output.png")
    parser.add_argument("--loglevel", metavar="LEVEL", default="INFO",
                        help="Logging level, default INFO.")
    parser.add_argument("--size", metavar="WxH", type=str, default=None,
                        help="Size of the screen. default None")
    parser.add_argument('--rasterizer', metavar='TYPE', type=str,
                        default="chromium", help='Rasterizer type.')

    args = parser.parse_args()
    logging.basicConfig(level=getattr(logging, args.loglevel.upper()))
    size = tuple(args.size.split("x")) if args.size else None
    rasterizer = create_rasterizer(args.rasterizer)
    image = rasterizer.rasterize(args.input, size=size)
    image.save(args.output)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
Inkscape rasterizer module.

Prerequisite:

    sudo apt-get install -y inkscape

"""
from __future__ import absolute_import, unicode_literals

from PIL import Image
import logging
import os
import sys
import subprocess
from psd2svg.utils import temporary_directory
from psd2svg.rasterizer.base_rasterizer import BaseRasterizer

logger = logging.getLogger(__name__)


class InkscapeRasterizer(BaseRasterizer):
    """Inkscape rasterizer."""

    def __init__(self, executable_path="inkscape", **kwargs):
        self.executable_path = executable_path

    def rasterize(self, url, size=None, format="png"):
        with temporary_directory() as tempdir:
            output_file = os.path.join(tempdir, "output.{}".format(format))
            cmd = [os.path.abspath(url), "-e", output_file,
                   "-b" "FFFFFF", "-y", "0"]
            if size:
                cmd += ["-w", size[0], "-h", size[1]]
            proc = subprocess.check_call(
                [self.executable_path, "-z"] + cmd,
                stdout=sys.stdout, stderr=sys.stdout)
            assert os.path.exists(output_file)
            rasterized = Image.open(output_file)
            return self.composite_background(rasterized)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
from logging import getLogger
import numpy as np
import svgwrite
from psd_tools.constants import TaggedBlock
from psd_tools.decoder.actions import List, Descriptor
from psd_tools.user_api.actions import Color
from psd2svg.converter.constants import BLEND_MODE
from psd2svg.utils.color import cmyk2rgb


logger = getLogger(__name__)


class EffectsConverter(object):

    def add_effects(self, layer, element):
        """
        Add effects to the element.

        Effects wraps the given element in a group, and the group will contain
        multiple items each corresponding to the applied effect.
        """

        fill_opacity = layer.get_tag(
            TaggedBlock.BLEND_FILL_OPACITY, 255) / 255.0
        if len(layer.effects) == 0:
            if fill_opacity < 1.0:
                element['opacity'] = layer.opacity / 255.0 * fill_opacity
            return element
        else:
            container = self._dwg.g()
            container['class'] = 'psd-effects'
            # Move class annotation.
            if 'class' in element.attribs:
                container['class'] += ' ' + element.attribs['class']
                del element.attribs['class']
            # Move subclass of title nodes.
            for child in element.elements:
                if isinstance(child, svgwrite.base.Title):
                    container.add(child)
            for child in container.elements:
                element.elements.remove(child)
            self._dwg.defs.add(element)


        # Outer effects.
        for effect in layer.effects.find('dropshadow'):
            shadow = self.create_drop_shadow(layer, effect, element)
            container.add(shadow)
        for effect in layer.effects.find('outerglow'):
            glow = self.create_outer_glow(layer, effect, element)
            container.add(glow)

        blend_mode = None  # Interior blend mode.
        if layer.get_tag(TaggedBlock.BLEND_INTERIOR_ELEMENTS, False):
            blend_mode = layer.blend_mode

        # Add the original.
        use = self._dwg.use(element.get_iri(), opacity=fill_opacity)
        use['class'] = 'layer-effect original'
        container.add(use)

        # Overlay effects.
        if layer.effects.has((
                'coloroverlay', 'patternoverlay', 'gradientoverlay')):

            mask = self._create_overlay_mask(layer, element)
            for effect in layer.effects.find('patternoverlay'):
                overlay = self.create_pattern_overlay(
                    layer, effect, mask, blend_mode)
                container.add(overlay)

            for effect in layer.effects.find('gradientoverlay'):
                overlay = self.create_gradient_overlay(
                    layer, effect, mask, blend_mode)
                container.add(overlay)

            for effect in layer.effects.find('coloroverlay'):
                overlay = self.create_color_overlay(
                    layer, effect, mask, blend_mode)
                container.add(overlay)

        # Inner effects.
        for effect in layer.effects.find('innershadow'):
            shadow = self.create_inner_shadow(
                layer, effect, element, blend_mode)
            container.add(shadow)

        for effect in layer.effects.find('innerglow'):
            glow = self.create_inner_glow(layer, effect, element, blend_mode)
            container.add(glow)

        # Bevel and emboss.
        for effect in layer.effects.find('bevelemboss'):
            bevelemboss = self.create_bevel_emboss(layer, effect, element)
            container.add(bevelemboss)

        # Satin.
        for effect in layer.effects.find('satin'):
            satin = self.create_satin(layer, effect, element)
            container.add(satin)

        # Stroke.
        for effect in layer.effects.find('stroke'):
            stroke = self.create_stroke(layer, effect, element)
            container.add(stroke)

        return container

    def _create_overlay_mask(self, layer, element):
        """Create a mask used for overlay effects."""
        if not layer.has_box():
            logger.warning('Fill effect to empty layer.')

        mask_element = self._dwg.defs.add(self._dwg.mask(
            size=(layer.width, layer.height)))
        mask_element['color-interpolation'] = 'sRGB'
        use = mask_element.add(self._dwg.use(element.get_iri()))
        use['filter'] = self._get_white_filter().get_funciri()
        return mask_element

    def create_color_overlay(self, layer, effect, mask, blend_mode):
        """Create a color overlay."""
        element = self._dwg.rect(
            size=(layer.width, layer.height),
            insert=(layer.left, layer.top),
            fill=self.create_solid_color(effect),
            mask=mask.get_funciri())
        self._add_overlay_attribute(
            element, effect, blend_mode, 'color-overlay')
        return element

    def create_pattern_overlay(self, layer, effect, mask, blend_mode):
        """Create a pattern overlay."""
        pattern = self.create_pattern(effect, (layer.left, layer.top))
        element = self._dwg.rect(
            size=(layer.width, layer.height),
            insert=(layer.left, layer.top),
            fill=pattern.get_funciri(),
            mask=mask.get_funciri())
        self._add_overlay_attribute(
            element, effect, blend_mode, 'pattern-overlay')
        return element

    def create_gradient_overlay(self, layer, effect, mask, blend_mode):
        """Create a gradient overlay."""
        gradient = self.create_gradient(effect, (layer.width, layer.height))
        element = self._dwg.rect(
            size=(layer.width, layer.height),
            insert=(layer.left, layer.top),
            fill=gradient.get_funciri(),
            mask=mask.get_funciri())
        self._add_overlay_attribute(
            element, effect, blend_mode, 'gradient-overlay')
        return element

    def _add_overlay_attribute(self, element, effect, blend_mode, kind):
        """Add common overlay attributes."""
        element['class'] = 'layer-effect {}'.format(kind)
        opacity = effect.opacity.value / 100.0  # TODO: Check unit
        if opacity < 1.0:
            element['fill-opacity'] = opacity
        self.add_blend_mode(
            element, blend_mode if blend_mode else effect.blend_mode)

    def create_drop_shadow(self, layer, effect, element):
        """Create a drop shadow effect."""
        blur = effect.size.value
        spread = effect.choke.value / 100.0
        angle = effect.angle.value
        radius = effect.distance.value
        dx = -radius * np.cos(np.radians(angle))
        dy = radius * np.sin(np.radians(angle))
        filt = self._dwg.defs.add(self._dwg.filter(
            x='-50%', y='-50%', size=('200%', '200%')))
        filt['class'] = 'drop-shadow'

        filt.feOffset('SourceAlpha', dx=dx, dy=dy, result='drshOffset')
        filt.feGaussianBlur('drshOffset', stdDeviation=blur / 2.0,
                            result='drshBlur',
                            **{'color-interpolation-filters': 'sRGB'})
        transfer = filt.feComponentTransfer('drshBlur', result='drshBlurA')
        transfer.feFuncA('linear', slope=1.0 + 4 * spread, intercept=0.0)
        flood = filt.feFlood(result='drshFlood')
        flood['flood-color'] = self.create_solid_color(effect)
        flood['flood-opacity'] = effect.opacity.value / 100.0
        filt.feComposite('drshFlood', in2='drshBlurA', operator='in',
                         result='drshShadow')

        shadow = self._dwg.use(element.get_iri(), filter=filt.get_funciri())
        shadow['class'] = 'layer-effect dropshadow'
        self.add_blend_mode(shadow, effect.blend_mode)
        return shadow

    def create_outer_glow(self, layer, effect, element):
        """Create an outer glow effect."""
        blur = effect.size.value
        spread = effect.choke.value / 100.0
        filt = self._dwg.defs.add(self._dwg.filter(
            x='-50%', y='-50%', size=('200%', '200%')))
        filt['class'] = 'outerglow'

        # Saturate alpha mask before glow if non-zero spread.
        if spread > 0:
            transfer = filt.feComponentTransfer(
                'SourceAlpha', result='orglAlpha')
            transfer.feFuncA('linear', slope=255, intercept=0.0)
            result = 'orglDilate'
            filt.feMorphology(
                'orglAlpha', radius=blur * spread, operator='dilate',
                result=result)
        else:
            result = 'SourceAlpha'
        filt.feGaussianBlur(
            result, stdDeviation=blur * (1 - spread), result='orglBlur',
            **{'color-interpolation-filters': 'sRGB'})
        transfer = filt.feComponentTransfer('orglBlur', result='orglBlurA')
        transfer.feFuncA('linear', slope=1 + 4 * spread, intercept=0.0)
        flood = filt.feFlood(result='orglFlood')

        if effect.color:
            flood['flood-color'] = self.create_solid_color(effect)
        else:
            logger.warning("Gradient glow not implemented")
            flood['flood-color'] = self.create_solid_color(
                effect.gradient.colors[0])
        flood['flood-opacity'] = effect.opacity.value / 100.0
        filt.feComposite('orglFlood', in2='orglBlurA', operator='in',
                         result='orglShadow')
        filt.feComposite('orglShadow', in2='SourceAlpha', operator='out',
                         result='orglShadowA')

        glow = self._dwg.use(element.get_iri(), filter=filt.get_funciri())
        glow['class'] = 'layer-effect outer-glow'
        self.add_blend_mode(glow, effect.blend_mode)
        return glow

    def create_inner_shadow(self, layer, effect, element, blend_mode):
        """Create inner shadow effect."""
        blur = effect.size.value
        angle = effect.angle.value
        radius = effect.distance.value
        dx = -radius * np.cos(np.radians(angle))
        dy = radius * np.sin(np.radians(angle))

        filt = self._dwg.defs.add(self._dwg.filter())
        filt['class'] = 'inner-shadow'
        flood = filt.feFlood(result='irshFlood')
        flood['flood-color'] = self.create_solid_color(effect)
        flood['flood-opacity'] = effect.opacity.value / 100.0
        filt.feComposite('irshFlood', in2='SourceAlpha', operator='out',
                         result='irshShadow')
        filt.feOffset('irshShadow', dx=dx, dy=dy, result='irshOffset')
        filt.feGaussianBlur('irshOffset', stdDeviation=blur / 2.0,
                            result='irshBlur',
                            **{'color-interpolation-filters': 'sRGB'})
        filt.feComposite('irshBlur', in2='SourceAlpha', operator='in',
                         result='irshShadow')

        shadow = self._dwg.use(element.get_iri(), filter=filt.get_funciri())
        shadow['class'] = 'layer-effect inner-shadow'
        self.add_blend_mode(
            shadow, blend_mode if blend_mode else effect.blend_mode)
        return shadow

    def create_inner_glow(self, layer, effect, element, blend_mode):
        """Create inner glow effect."""
        blur = effect.size.value
        spread = effect.choke.value / 100.0

        # Real inner glow needs distance transform.
        filt = self._dwg.defs.add(self._dwg.filter())
        filt['class'] = 'inner-glow'
        flood = filt.feFlood(result='irglFlood')
        # TODO: Gradient fill
        if effect.color:
            flood['flood-color'] = color = self.create_solid_color(effect)
        else:
            logger.warning("Gradient glow not implemented")
        flood['flood-opacity'] = effect.opacity.value / 100.0
        # Saturate alpha mask before glow.
        transfer = filt.feComponentTransfer('SourceAlpha', result='irglAlpha')
        transfer.feFuncA('linear', slope=255, intercept=0)
        filt.feComposite('irglFlood', in2='irglAlpha', operator='out',
                         result='irglShadow')
        filt.feMorphology('irglShadow', radius=blur * spread,
                          operator='dilate', result='irglDilate')
        filt.feGaussianBlur('irglDilate', stdDeviation=blur * (1 - spread),
                            result='irglBlur',
                            **{'color-interpolation-filters': 'sRGB'})
        filt.feComposite('irglBlur', in2='irglAlpha', operator='in',
                         result='irglShadow')

        glow = self._dwg.use(element.get_iri(), filter=filt.get_funciri())
        glow['class'] = 'layer-effect inner-glow'
        self.add_blend_mode(
            glow, blend_mode if blend_mode else effect.blend_mode)
        return glow

    def create_bevel_emboss(self, layer, effect, element):
        """Create bevel and emboss effect."""
        # In SVG, bevel and emboss need to be split into two elements.

        # Shadow.
        filt = self._dwg.defs.add(self._dwg.filter())
        shadow = self._dwg.use(
            element.get_iri(), filter=filt.get_funciri())
        shadow['class'] = 'layer-effect bevel-emboss shadow'
        shadow['opacity'] = effect.shadow_opacity.value / 100.0
        self.add_blend_mode(shadow, effect.shadow_mode)
        filt['class'] = 'bevel-emboss shadow'
        if effect.bevel_style == 'inner-bevel':
            blur = filt.feGaussianBlur('SourceAlpha', result='blur',
                stdDeviation=effect.size.value / 2.0,
                **{'color-interpolation-filters': 'sRGB'})
            light = filt.feDiffuseLighting('blur',
                result='shadow',
                surfaceScale=effect.size.value / 2.0,
                diffuseConstant=2.0)
            light.feDistantLight(azimuth=-effect.angle.value,
                                 elevation=effect.altitude.value)
            color = [(x / 255.0) for x in effect.shadow_color.value]
            filt.feColorMatrix(
                'shadow',
                result='color-shadow',
                type='matrix',
                values="0 0 0 0 {:g} 0 0 0 0 {:g} "
                       "0 0 0 0 {:g} -1.0 0 0 0 1.0".format(*color))
            filt.feComposite('color-shadow', in2='SourceAlpha', operator='in')
        elif effect.bevel_style == 'emboss':
            filt['x'], filt['y'] = '-10%', '-10%'
            filt['width'], filt['height'] = '120%', '120%'
            blur = filt.feGaussianBlur('SourceAlpha', result='blur',
                stdDeviation=effect.size.value / 3.2,
                **{'color-interpolation-filters': 'sRGB'})
            light = filt.feDiffuseLighting('blur',
                result='shadow',
                surfaceScale=effect.size.value / 1.2,
                diffuseConstant=2.0)
            light.feDistantLight(azimuth=-effect.angle.value,
                                 elevation=effect.altitude.value)
            color = [(x / 255.0) for x in effect.shadow_color.value]
            filt.feColorMatrix(
                'shadow',
                result='color-shadow',
                type='matrix',
                values="0 0 0 0 {:g} 0 0 0 0 {:g} "
                       "0 0 0 0 {:g} -1.0 0 0 0 1.0".format(*color))
        else:
            logger.warning("Not implemented: {}".format(effect.bevel_style))

        # Highlight.
        filt = self._dwg.defs.add(self._dwg.filter())
        highlight = self._dwg.use(
            element.get_iri(), filter=filt.get_funciri())
        highlight['class'] = 'layer-effect bevel-emboss highlight'
        highlight['opacity'] = effect.highlight_opacity.value / 100.0
        self.add_blend_mode(highlight, effect.highlight_mode)
        filt['class'] = 'bevel-emboss highlight'
        if effect.bevel_style == 'inner-bevel':
            blur = filt.feGaussianBlur('SourceAlpha', result='blur',
                stdDeviation=effect.size.value / 2.0,
                **{'color-interpolation-filters': 'sRGB'})
            light = filt.feSpecularLighting('blur',
                result='highlight',
                surfaceScale=effect.size.value / 1.6,
                specularExponent=14.0,
                specularConstant=1.0)
            light.feDistantLight(azimuth=-effect.angle.value,
                                 elevation=effect.altitude.value)
            light['lighting-color'] = self.create_solid_color(
                effect.highlight_color)
            filt.feComposite('highlight', in2='SourceAlpha', operator='in')
        elif effect.bevel_style == 'emboss':
            filt['x'], filt['y'] = '-10%', '-10%'
            filt['width'], filt['height'] = '120%', '120%'
            blur = filt.feGaussianBlur('SourceAlpha', result='blur',
                stdDeviation=effect.size.value / 3.2,
                **{'color-interpolation-filters': 'sRGB'})
            light = filt.feSpecularLighting('blur',
                result='highlight',
                surfaceScale=effect.size.value / 3.2,
                specularExponent=14.0,
                specularConstant=1.0)
            light.feDistantLight(azimuth=-effect.angle.value,
                                 elevation=effect.altitude.value)
            light['lighting-color'] = self.create_solid_color(
                effect.highlight_color)
        else:
            logger.warning("Not implemented: {}".format(effect.bevel_style))

        container = self._dwg.g()
        container.add(shadow)
        container.add(highlight)
        container['class'] = 'layer-effect bevel-emboss'
        return container

    def create_satin(self, layer, effect, element):
        """Create satin effect."""

        """
        Sating effect is complicated to reproduce:

        1. Create two shifted self-shapes from two sides, where the
           shift arrangement is determined from the angle parameter and
           distance. The two shape masks should follow even-odd filling rule.
        2. Invert the filled area if inverted.
        3. Apply Gaussian blur to the filled area from the two rectangles.
        4. For the filled area, apply specified color and opacity.
        """
        angle = effect.angle.value
        radius = effect.distance.value
        dx = -radius * np.cos(np.radians(angle))
        dy = radius * np.sin(np.radians(angle))

        filt = self._dwg.defs.add(self._dwg.filter())
        filt['class'] = 'satin'
        filt.feOffset('SourceAlpha', result='shape1', dx=dx, dy=dy)
        filt.feOffset('SourceAlpha', result='shape2', dx=-dx, dy=-dy)
        filt.feComposite('shape1', in2='shape2', result='xor', operator='xor')
        color = [x / 255.0 for x in effect.color.value]
        if effect.inverted:
            filt.feColorMatrix(
                'xor',
                result='xor-shaded',
                type='matrix',
                values='1 0 0 0 {:g} '
                       '0 1 0 0 {:g} '
                       '0 0 1 0 {:g} '
                       '0 0 0 -1 1 '.format(*color),
                )
        else:
            filt.feColorMatrix(
                'xor',
                result='xor-shaded',
                type='matrix',
                values='1 0 0 0 {:g} '
                       '0 1 0 0 {:g} '
                       '0 0 1 0 {:g} '
                       '0 0 0 1 0 '.format(*color),
                )
        filt.feGaussianBlur('xor-shaded', result='blur',
            stdDeviation=effect.size.value / 2.0,
            **{'color-interpolation-filters': 'sRGB'})
        filt.feComposite('blur', in2='SourceAlpha', operator='in')

        satin = self._dwg.use(
            element.get_iri(), filter=filt.get_funciri())
        satin['class'] = 'layer-effect satin'
        satin['opacity'] = effect.opacity.value / 100.0

        self.add_blend_mode(satin, effect.blend_mode)
        return satin

    def create_stroke(self, layer, effect, element):
        """Create a stroke effect."""
        radius = int(effect.size.value)  # TODO: Check unit.
        style = effect.position

        filt = self._dwg.defs.add(self._dwg.filter())
        filt['class'] = 'stroke'
        flood = filt.feFlood(result='frfxFlood')
        # TODO: Implement gradient or pattern fill
        if effect.fill_type == 'solid-color':
            flood['flood-color'] = self.create_solid_color(effect.fill)
        elif effect.fill_type == 'pattern':
            logger.warning("Pattern stroke not implemented")
        elif effect.fill_type == 'gradient':
            logger.warning("Gradient stroke not implemented")
            flood['flood-color'] = self.create_solid_color(
                effect.fill.gradient.colors[0])
        else:
            logger.warning("Unknown fill type: {}".format(effect.fill_type))


        flood['flood-opacity'] = effect.opacity.value / 100.0
        if style == 'outer':
            filt.feMorphology('SourceAlpha', result='frfxMorph',
                              operator='dilate', radius=radius)
            filt.feComposite('frfxFlood', in2='frfxMorph', operator='in',
                             result='frfxBoundary')
            filt.feComposite('frfxBoundary', in2='SourceAlpha',
                             operator='out')
        elif style == 'inner':
            filt.feMorphology('SourceAlpha', result='frfxMorph',
                              operator='erode', radius=radius)
            filt.feComposite('frfxFlood', in2='frfxMorph', operator='out',
                             result='frfxBoundary')
            filt.feComposite('frfxBoundary', in2='SourceAlpha',
                             operator='in')
        else:
            filt.feMorphology('SourceAlpha', result='frfxDilate',
                              operator='dilate', radius=radius / 2.0)
            filt.feMorphology('SourceAlpha', result='frfxErode',
                              operator='erode', radius=radius / 2.0)
            filt.feComposite('frfxDilate', in2='frfxErode', operator='out',
                             result='frfxMorph')
            filt.feComposite('frfxFlood', in2='frfxMorph', operator='in')

        stroke = self._dwg.use(element.get_iri(), filter=filt.get_funciri())
        stroke['class'] = 'layer-effect stroke'
        self.add_blend_mode(stroke, effect.blend_mode)
        return stroke

    def _get_white_filter(self, color='white'):
        if not self._white_filter:
            self._white_filter = self._dwg.defs.add(self._dwg.filter())
            self._white_filter['class'] = 'white-filter'
            if color == 'white':
                self._white_filter.feColorMatrix(
                    'SourceAlpha', type='matrix',
                    values="0 0 0 0 1 0 0 0 0 1 0 0 0 0 1 0 0 0 1 0")
            else:
                self._white_filter.feColorMatrix(
                    'SourceAlpha', type='matrix',
                    values="0 0 0 1 0 0 0 0 1 0 0 0 0 1 0 0 0 0 1 0")
        return self._white_filter

    def _get_identity_filter(self):
        if not self._identity_filter:
            self._identity_filter = self._dwg.defs.add(self._dwg.filter())
            self._identity_filter['class'] = 'identify-filter'
            transfer = self._identity_filter.feComponentTransfer(
                'SourceGraphic')
            transfer['color-interpolation'] = 'sRGB'
            transfer.feFuncR('identity')
            transfer.feFuncG('identity')
            transfer.feFuncB('identity')
            transfer.feFuncA('identity')
        return self._identity_filter



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
from importlib import import_module


def create_rasterizer(name="chromium", *args, **kwargs):
    module_name = "psd2svg.rasterizer.{}_rasterizer".format(name.lower())
    class_name = "{}Rasterizer".format(name.capitalize())
    cls = getattr(import_module(module_name), class_name)
    assert cls is not None, "Invalid class name: {}".format(class_name)
    return cls(*args, **kwargs)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
from logging import getLogger
import svgwrite
from psd2svg.converter.adjustments import AdjustmentsConverter
from psd2svg.converter.core import LayerConverter
from psd2svg.converter.effects import EffectsConverter
from psd2svg.converter.io import PSDReader, SVGWriter
from psd2svg.converter.shape import ShapeConverter
from psd2svg.converter.text import TextConverter
from psd2svg.version import __version__


logger = getLogger(__name__)


def psd2svg(input, output=None, **kwargs):
    converter = PSD2SVG(**kwargs)
    return converter.convert(input, output)


class PSD2SVG(AdjustmentsConverter, EffectsConverter, LayerConverter,
              PSDReader, ShapeConverter, SVGWriter, TextConverter):
    """PSD to SVG converter

    input_url - url, file-like object, PSDImage, or any of its layer.
    output_url - url or file-like object to export svg. if None, return data.
    text_mode - option to switch text rendering (default 'image')
      * 'image' use Photoshop's bitmap
      * 'text' use SVG text
      * 'image-only' use Photoshop's bitmap and discard SVG text
      * 'text-only' use SVG text and discard Photoshop bitmap
    export_resource - use dataURI to embed bitmap (default True)
    """
    def __init__(self, text_mode='image', export_resource=False,
                 resource_prefix='', overwrite=True, reset_id=True,
                 embed_preview=False):
        self.text_mode = text_mode
        self.export_resource = export_resource
        self.resource_prefix = resource_prefix
        self.overwrite = overwrite
        self.reset_id = reset_id
        self.embed_preview = embed_preview

    def reset(self):
        """Reset the converter."""
        self._psd = None
        self._input_layer = None
        self._white_filter = None
        self._identity_filter = None
        if self.reset_id:
            svgwrite.utils.AutoID._set_value(0)

    def convert(self, input, output=None, layer_view=True):
        """Convert the given PSD to SVG."""
        self.reset()
        self._load(input)
        self._set_output(output)

        if (not self.overwrite and self._output_file and
                self._output.exists(self._output_file)):
            url = self._output.url(self._output_file)
            logger.warning('File exists: {}'.format(url))
            return url

        if layer_view and self._input_layer and self._input_layer.has_box():
            self._dwg = svgwrite.Drawing(
                size=(self._input_layer.width, self._input_layer.height),
                viewBox="{} {} {} {}".format(
                    self._input_layer.left, self._input_layer.top,
                    self._input_layer.width, self._input_layer.height))
        else:
            self._dwg = svgwrite.Drawing(
                size=(self.width, self.height),
                viewBox="0 0 {} {}".format(self.width, self.height))

        # if self.text_mode in ('image', 'image-only'):
        #     stylesheet = '.text-text { display: none; }'
        # elif self.text_mode in ('text', 'text-only'):
        #     stylesheet = '.text-image { display: none; }'
        # self._dwg.defs.add(self._dwg.style(stylesheet))

        # Add layers.
        if self._input_layer:
            self._dwg.add(self.convert_layer(self._input_layer))
        else:
            self.create_group(self._psd, element=self._dwg)
            empty_psd = len(self._psd.layers) == 0
            if self.embed_preview or empty_psd:
                self._dwg.add(self.create_preview(hidden=not empty_psd))

        return self._save_svg()

    @property
    def width(self):
        return self._psd.width if hasattr(self, '_psd') else None

    @property
    def height(self):
        return self._psd.height if hasattr(self, '_psd') else None



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
from logging import getLogger
from psd_tools.constants import TaggedBlock, PathResource
from psd2svg.converter.constants import BLEND_MODE


logger = getLogger(__name__)


class ShapeConverter(object):

    def generate_path(self, vector_mask, command='C'):
        """Sequence generator for SVG path constructor."""

        # TODO: Implement even-odd rule for multiple paths.
        # first path --> show, second path --> hide, third path --> show.
        # should be clipPath.
        for path in vector_mask.paths:
            if path.num_knots == 0:
                continue

            # Initial point.
            yield 'M'
            yield path.knots[0].anchor[1] * self.width
            yield path.knots[0].anchor[0] * self.height
            yield command

            # Closed path or open path
            points = (zip(path.knots, path.knots[1:] + path.knots[0:1])
                      if path.closed else zip(path.knots, path.knots[1:]))

            # Rest of the points.
            for p1, p2 in points:
                yield p1.leaving_knot[1] * self.width
                yield p1.leaving_knot[0] * self.height
                yield p2.preceding_knot[1] * self.width
                yield p2.preceding_knot[0] * self.height
                yield p2.anchor[1] * self.width
                yield p2.anchor[0] * self.height

            if path.closed:
                yield 'Z'


    def add_stroke_style(self, layer, element):
        """Add stroke style to the path element."""
        if not layer.has_stroke():
            return element

        stroke = layer.stroke
        if not stroke.enabled:
            return element

        if stroke.line_alignment == 'inner':
            clippath = self._dwg.defs.add(self._dwg.clipPath())
            clippath['class'] = 'psd-stroke stroke-inner'
            clippath.add(self._dwg.path(
                self.generate_path(layer.vector_mask)))
            element['stroke-width'] = stroke.line_width.value * 2
            element['clip-path'] = clippath.get_funciri()
        # elif stroke.line_alignment == 'outer':
        #     mask = self._dwg.defs.add(self._dwg.mask())
        #     mask['class'] = 'psd-stroke stroke-outside'
        #     mask.add(self._dwg.rect(
        #         insert=(layer.left, layer.top),
        #         size=(layer.width, layer.height),
        #         fill='white'))
        #     mask.add(self._dwg.path(
        #         self.generate_path(layer.vector_mask), fill='black'))
        #     element['stroke-width'] = stroke.line_width * 2
        #     element['mask'] = mask.get_funciri()
        else:
            element['stroke-width'] = stroke.line_width.value
        # element['stroke-alignment'] = stroke.line_alignment

        if stroke.content.name == 'patternoverlay':
            pattern = self.create_pattern(
                stroke.content, insert=(layer.left, layer.top))
            element['stroke'] = pattern.get_funciri()
        elif stroke.content.name == 'gradientoverlay':
            gradient = self.create_gradient(
                stroke.content, size=(layer.width, layer.height))
            element['stroke'] = gradient.get_funciri()
        elif stroke.content.name == 'coloroverlay':
            element['stroke'] = self.create_solid_color(stroke.content)

        if not stroke.fill_enabled:
            element['fill-opacity'] = 0

        element['stroke-opacity'] = stroke.opacity.value / 100.0
        element['stroke-linecap'] =  stroke.line_cap_type
        element['stroke-linejoin'] = stroke.line_join_type
        if stroke.line_dash_set:
            element['stroke-dasharray'] = ",".join(
                [str(x.value * stroke.line_width.value)
                 for x in stroke.line_dash_set])
            element['stroke-dashoffset'] = stroke.line_dash_offset.value
        self.add_blend_mode(element, stroke.blend_mode)

        return element


    def add_stroke_content_style(self, layer, element):
        """Add stroke content (fill) style to the path element."""
        if not layer.has_stroke_content():
            return element

        effect = layer.stroke_content
        if not effect.enabled:
            return element

        if effect.name == 'patternoverlay':
            pattern = self.create_pattern(
                effect, insert=(layer.left, layer.top))
            element['fill'] = pattern.get_funciri()
        elif effect.name == 'gradientoverlay':
            gradient = self.create_gradient(
                effect, size=(layer.width, layer.height))
            element['fill'] = gradient.get_funciri()
        if effect.name == 'coloroverlay':
            element['fill'] = self.create_solid_color(effect)

        return element



# -*- coding: utf-8 -*-
"""
Chromium-based rasterizer module.

Prerequisite:

    sudo apt-get install -y chromedriver chromium

"""
from __future__ import absolute_import, unicode_literals

from selenium import webdriver
from PIL import Image
from io import BytesIO
import logging
import math
import os
import re
import sys
import xml.etree.ElementTree as ET
from psd2svg.rasterizer.base_rasterizer import BaseRasterizer

logger = logging.getLogger(__name__)


# CHROMEDRIVER_PATH = "/usr/lib/chromium-browser/chromedriver"
VIEWPORT_SIZE = (16, 16)  # Default size when nothing is specified.


class ChromiumRasterizer(BaseRasterizer):
    """Chromium rasterizer."""

    def __init__(self, executable_path="chromedriver", dpi=96.0, **kwargs):
        options = webdriver.ChromeOptions()
        options.add_argument("headless")
        options.add_argument("disable-gpu")
        options.add_argument("disable-infobars")
        options.add_argument("enable-experimental-web-platform-features")
        options.add_argument("default-background-color FFFFFF00")
        self.driver = webdriver.Chrome(
            executable_path=executable_path,
            chrome_options=options)
        self.dpi = dpi

    def __del__(self):
        self.driver.quit()

    def rasterize(self, url, size=None):
        if not re.match(r"^(\S+)://.*$", url):
            url = "file://" + os.path.abspath(url)
        if size:
            self.driver.set_window_size(*size)
            self.driver.get(url)
        else:
            self.driver.get(url)
            size = self._set_windowsize()
        rasterized = Image.open(BytesIO(self.driver.get_screenshot_as_png()))
        if rasterized.width != size[0] or rasterized.height != size[1]:
            logger.info("Resizing captured screenshot from {} to {}".format(
                rasterized.size, size))
            rasterized = rasterized.resize(size, Image.NEAREST)
        return self.composite_background(rasterized)

    def _set_windowsize(self):
        svg = self.driver.find_element_by_tag_name("svg")
        if not svg:
            return
        width = svg.get_attribute("width")
        height = svg.get_attribute("height")
        if not width or not height:
            return
        logger.debug("Resizing to {}x{}".format(width, height))
        width = self._get_pixels(width)
        height = self._get_pixels(height)
        if width == 0 or height == 0:
            width, height = VIEWPORT_SIZE
        self.driver.set_window_size(int(width), int(height))
        return width, height

    def _get_pixels(self, value):
        match = re.match(r"(?P<value>\d+)(?P<unit>\D+)?", value)
        value = int(match.group("value"))
        unit = match.group("unit")
        if unit == "pt":
            value = math.ceil(value * self.dpi / 72.0)
        return value



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
import base64
import hashlib
import io
from logging import getLogger
import os
from psd_tools.user_api.psd_image import PSDImage
from psd_tools.user_api.layers import _RawLayer
from psd2svg.storage import get_storage
import xml.dom.minidom as minidom


logger = getLogger(__name__)


class PSDReader(object):

    def _load(self, input_data):
        if hasattr(input_data, "read"):
            self._load_stream(input_data)
            self._input_layer = None
        elif isinstance(input_data, PSDImage):
            self._load_psd(input_data)
            self._input_layer = None
        elif isinstance(input_data, _RawLayer):
            self._load_psd(input_data._psd)
            self._input_layer = input_data
        else:
            self._load_storage(input_data)
            self._input_layer = None

    def _load_storage(self, url):
        storage = get_storage(os.path.dirname(url))
        filename = os.path.basename(url)
        logger.info('Opening {}'.format(url))
        with storage.open(filename) as f:
            self._load_stream(f)
        self._input = url

    def _load_stream(self, stream):
        self._input = None
        self._psd = PSDImage.from_stream(stream)

    def _load_psd(self, psd):
        self._input = None
        self._psd = psd


class SVGWriter(object):

    def _set_output(self, output_data):
        # IO object.
        if not output_data:
            self._output = None
            self._resource = get_storage(self.resource_prefix)
            self._output_file = None
            return
        if hasattr(output_data, 'write'):
            self._output = output_data
            self._resource = get_storage(self.resource_prefix)
            self._output_file = None
            return

        # Else save to a file.
        if not output_data.endswith('/') and os.path.isdir(output_data):
            output_data += '/'
        self._output = get_storage(os.path.dirname(output_data))
        self._resource = get_storage(
            self._output.url(self.resource_prefix))
        self._output_file = os.path.basename(output_data)
        if not self._output_file:
            if self._input:
                basename = os.path.splitext(os.path.basename(self._input))[0]
                self._output_file = basename + '.svg'
            else:
                raise ValueError('Invalid output: {}'.format(output_data))

    def _save_svg(self):
        # Write to the output.
        pretty_string = self._get_svg()
        if self._output_file:
            url = self._output.url(self._output_file)
            logger.info('Saving {}'.format(url))
            self._output.put(self._output_file, pretty_string.encode('utf-8'))
            return url
        elif self._output:
            self._output.write(pretty_string)
            return self._output
        else:
            return pretty_string

    def _get_svg(self):
        with io.StringIO() as f:
            # svgwrite's pretty option is not compatible with Python 2.7
            # unicode. Here we manually encode utf-8.
            self._dwg.write(f, pretty=False)
            xml_string = f.getvalue().encode('utf-8')

        xml_tree = minidom.parseString(xml_string)
        return xml_tree.toprettyxml(indent='  ')

    def _get_image_href(self, image, fmt='png', icc_profile=None):
        output = io.BytesIO()
        image.save(output, format=fmt, icc_profile=icc_profile)
        encoded_image = output.getvalue()
        output.close()
        if self.export_resource:
            checksum = hashlib.md5(encoded_image).hexdigest()
            filename = checksum + '.' + fmt
            if self.overwrite or not self._resource.exists(filename):
                logger.info('Saving {}'.format(
                    self._resource.url(filename)))
                self._resource.put(filename, encoded_image)
            href = os.path.join(self.resource_prefix, filename)
        else:
            href = ('data:image/{};base64,'.format(fmt) +
                    base64.b64encode(encoded_image).decode('utf-8'))
        return href
